include winim/[inc\winuser]
import re, strformat, strutils
import wNim/[wApp, wFrame]

let           
  nsiApp* = App() 
  nsiFrame* = Frame(title="Nim-SciTE interface", size=(320,220))

proc MyFindWindow*(class = "", title = ""): HWND = 
  let 
    className = newWideCString(class, class.len)
    titleName = newWideCString(title, title.len)
  result = FindWindowW(if class.len > 0: cast[LPWSTR](className) else: nil,
                       if title.len > 0: cast[LPWSTR](titleName) else: nil)

type
  WndID = object
    pid, tid: int
  EnumWnd = object
    hwnd: HWND
    class: WideCStringObj
    title: WideCStringObj
    pid, tid: int

var 
  sEnumWnd = newSeq[EnumWnd]()
  lp: LPARAM
  classfilter = newWideCString("", 256)
  titlefilter = newWideCString("", 4096)


proc my_WndEnumProc(hwnd: HWND, lp: LPARAM): WINBOOL {.stdcall.} =
  if GetParent(hwnd) == 0:
    var buflen = GetWindowTextLengthW(hwnd) + 1
    if buflen == 0: return TRUE

    var buf = newWideCString("", buflen)
    discard GetWindowTextW(hwnd, cast[LPWSTR](addr buf[0]), buflen)
    if titlefilter.len > 0:
      if fmt"{buf}" != fmt"{titlefilter}": return TRUE

    var className = newWideCString("", 4096)
    discard GetClassNameW(hwnd, cast[LPWSTR](className), 4096)
    if classfilter.len > 0:
      if fmt"{className}" != fmt"{classfilter}": return TRUE

    var
      pPID: LPDWORD
      pTID = GetWindowThreadProcessId(hwnd, cast[LPDWORD](addr pPID))
      oWndID = WndID(pid: cast[int](pPID), tid: cast[int](pTID))

    sEnumWnd.add(EnumWnd(hwnd: hwnd, 
                        class: className, 
                        title: buf,
                        pid: oWndID.pid,
                        tid: oWndID.tid))
  return TRUE


proc GetIDByWindow(hWnd: HWND): WndID =
  var 
    pPID: LPDWORD
    pTID = GetWindowThreadProcessId(hWnd, cast[LPDWORD](addr pPID))
  result = WndID(pid: cast[int](pPID), tid: cast[int](pTID))


proc GetEnumWindows(cfilter, tfilter: string): int =
  titlefilter = newWideCString(tfilter, tfilter.len + 1)
  classfilter = newWideCString(cfilter, cfilter.len + 1)
  sEnumWnd = newSeq[EnumWnd]()
  EnumWindows(my_WndEnumProc, lp)
  result = sEnumWnd.len


proc GetHwndDirectorExtension(): HWND =
  # let hActive = GetForegroundWindow()         # zum Test auskommentiert
  let hScite = MyFindWindow("SciTEWindow")
  # if hActive != hScite: return 0              # zum Test auskommentiert
  let pidScite = GetIDByWindow(hScite).pid
  let dirExtCount = GetEnumWindows("DirectorExtension", "")
  if dirExtCount == 1: result = sEnumWnd[0].hwnd
  else:
    for i in 0..dirExtCount - 1:
      if pidScite == sEnumWnd[i].pid: return sEnumWnd[i].hwnd


# ============================================================================ #

let hRecv* = nsiFrame.handle
var 
  received*: string  # enthält Ergebnis einer Abfrage
  pattern: Regex

nsiFrame.connect(WM_COPYDATA) do (event: wEvent):
  var 
    cds: PCOPYDATASTRUCT = cast[PCOPYDATASTRUCT](event.lParam)
    len: int = cds.cbData
    ret: cstring = cast[cstring](cds.lpData)
  received = $ret
  received.setLen(len)
  received = received.replace(pattern)


proc SendCommand(cmd: string, recv: bool = false): bool {.discardable.} =
  # let hScite = MyFindWindow(class = "DirectorExtension")
  let hScite = GetHwndDirectorExtension()
  var
    cmdScite = ":" & $hRecv & ":" & cmd
    cds: COPYDATASTRUCT
  cds.dwData = 1
  cds.cbData = DWORD(cmdScite.len + 1)
  cds.lpData = cmdScite[0].addr
  received = ""
  SendMessage(hScite, WM_COPYDATA, cast[WPARAM](hRecv), cast[LPARAM](cds.addr))
  return true

# ===== EDITOR procedures =================================================== # 

type
  EditorPosition* = tuple   # type for LuaCurrPos
    pos: int        # current position in editor, 0 based
    line: int       # current line number, 0 based
    content: string # content of current line w.o. line break
    length: int     # length of the content
    home: int       # home position of current line
    asc: int        # ASCII value from character on current position
    chr: string     # character from current position
    col: int        # column in current line, 0 based
    style: int      # style on current position



let LuaCurrPos* = """function currPos(_pos) local sep, result, tmp, tmp1, tmp2, tmp3 = '<@@>', '', _pos or editor.CurrentPos result = tostring(tmp)..sep  tmp1 = editor:LineFromPosition(tmp) result = result..tostring(tmp1)..sep  tmp2 = editor:GetLine(tmp1)""" & 
r"tmp2 = string.gsub(tmp2, '[\\r\\n]+$', '') " & """
result = result..tostring(tmp2)..sep..tostring(#tmp2)..sep  tmp1 = editor:PositionFromLine(tmp1) result = result..tostring(tmp1)..sep tmp1 = editor.CharAt[tmp] result = result..tostring(tmp1)..sep tmp1 = string.char(tmp1) result = result..tostring(tmp1)..sep tmp1 = editor.Column[tmp] result = result..tostring(tmp1)..sep tmp1 = editor.StyleAt[tmp] result = result..tostring(tmp1) return result end """



proc CurrentFile*(): bool {.discardable.} =
  pattern = re"^:\d+:[^:]+:"
  result = SendCommand("askfilename:", true)

proc GetProperty*(property: string): bool {.discardable.} =
  pattern = re"^:\d+:([^:]+:){2}"
  result = SendCommand("askproperty:" & property, true)

proc SetProperty*(property, value: string): bool {.discardable.} =
  result = SendCommand("property:" & property & "=" & value)

proc OutputConsole*(value: string): bool {.discardable.} =
  result = SendCommand("output:" & value)

proc MenuCmd*(cmdID: int): bool {.discardable.} =
  result = SendCommand("menucommand:" & $cmdID)

proc ClearOutput*(): bool {.discardable.} =
  result = SendCommand("extender:dostring do output:ClearAll() end")

proc LuaFuncRun*(lua: string, params: string = ""): bool {.discardable.} =
  var 
    match: array[2,string]
    fun = lua.match(re"^(?:local )?function ([^\(]+)\(([^\)]+)?", match)
  if not fun:
    fun = lua.match(re"^(?:local\s)?(\w+)\s*=\s*function\(([^\)]+)?", match) 
    if not fun: return false
  SendCommand(r"extender:dostring " & lua & r" props['extender.result']=" & match[0] & "(" & params & ")")

proc ParseLuaCurrPos*(): EditorPosition =
  let 
    sep = "<@@>"
    split = received.split(sep)
  # echo "received: ", received
  var tep: EditorPosition
  if split.len != 9: return tep
  tep[0] = parseInt(split[0])
  tep[1] = parseInt(split[1])
  tep[2] = $(split[2])
  tep[3] = parseInt(split[3])
  tep[4] = parseInt(split[4])
  tep[5] = parseInt(split[5])
  tep[6] = $(split[6])
  tep[7] = parseInt(split[7])
  tep[8] = parseInt(split[8])
  result = tep
  SetProperty("extender.result", "")

proc LuaFuncGet*(fLua: string, fParse: proc, T: type, param: string = ""): T =  
  LuaFuncRun(fLua, param)
  GetProperty("extender.result")
  result = fParse()

